@@ -344,7 +344,7 @@ plines_win(
     linenr_T	lnum,
     int		limit_winheight)	// when TRUE limit to window height
 {
-#if defined(FEAT_DIFF) || defined(PROTO)
+#if defined(FEAT_DIFF)
     // Check for filler lines above this buffer line.  When folded the result
     // is one line anyway.
     return plines_win_nofill(wp, lnum, limit_winheight)
@@ -565,6 +565,26 @@ gchar_cursor(void)
     return (int)*ml_get_cursor();
 }
 
+/*
+ * Return the character immediately before the cursor.
+ */
+    int
+char_before_cursor(void)
+{
+    if (curwin->w_cursor.col == 0)
+	return -1;
+
+    char_u *line = ml_get_curline();
+
+    if (has_mbyte)
+    {
+	char_u *p = line + curwin->w_cursor.col;
+	int prev_len = (*mb_head_off)(line, p - 1) + 1;
+	return mb_ptr2char(p - prev_len);
+    }
+    return line[curwin->w_cursor.col - 1];
+}
+
 /*
  * Write a character at the current cursor position.
  * It is directly written into the block.
@@ -655,7 +675,7 @@ ask_yesno(char_u *str, int direct)
     return r;
 }
 
-#if defined(FEAT_EVAL) || defined(PROTO)
+#if defined(FEAT_EVAL)
 
 /*
  * Returns the current mode as a string in "buf[MODE_MAX_LENGTH]", NUL
@@ -851,6 +871,7 @@ get_keystroke(void)
     int		save_mapped_ctrl_c = mapped_ctrl_c;
     int		waited = 0;
 
+    mod_mask = 0;
     mapped_ctrl_c = FALSE;	// mappings are not used here
     for (;;)
     {
@@ -954,7 +975,7 @@ get_keystroke(void)
     vim_free(buf);
 
     mapped_ctrl_c = save_mapped_ctrl_c;
-    return n;
+    return merge_modifyOtherKeys(n, &mod_mask);
 }
 
 // For overflow detection, add a digit safely to an int value.
@@ -996,7 +1017,7 @@ get_number(
     ++allow_keys;		// no mapping here, but recognize keys
     for (;;)
     {
-	windgoto(msg_row, msg_col);
+	windgoto(msg_row, cmdline_col_off + msg_col);
 	c = safe_vgetc();
 	if (VIM_ISDIGIT(c))
 	{
@@ -1329,7 +1350,7 @@ init_homedir(void)
     }
 }
 
-#if defined(EXITFREE) || defined(PROTO)
+#if defined(EXITFREE)
     void
 free_homedir(void)
 {
@@ -1343,7 +1364,7 @@ free_users(void)
 }
 #endif
 
-#if defined(MSWIN) || defined(PROTO)
+#if defined(MSWIN)
 /*
  * Initialize $VIM and $VIMRUNTIME when 'enc' is updated.
  */
@@ -2024,7 +2045,7 @@ vim_unsetenv_ext(char_u *var)
 	didset_vimruntime = FALSE;
 }
 
-#if defined(FEAT_EVAL) || defined(PROTO)
+#if defined(FEAT_EVAL)
 /*
  * Set environment variable "name" and take care of side effects.
  */
@@ -2262,7 +2283,7 @@ prepare_to_exit(void)
     else
 #endif
     {
-	windgoto((int)Rows - 1, 0);
+	windgoto((int)Rows - 1, cmdline_col_off);
 
 	/*
 	 * Switch terminal mode back now, so messages end up on the "normal"
@@ -2352,7 +2373,7 @@ fast_breakcheck(void)
     }
 }
 
-# if defined(FEAT_SPELL) || defined(PROTO)
+# if defined(FEAT_SPELL)
 /*
  * Like line_breakcheck() but check 100 times less often.
  */
@@ -2368,8 +2389,7 @@ veryfast_breakcheck(void)
 #endif
 
 #if defined(VIM_BACKTICK) || defined(FEAT_EVAL) \
-	|| (defined(HAVE_LOCALE_H) || defined(X_LOCALE)) \
-	|| defined(PROTO)
+	|| (defined(HAVE_LOCALE_H) || defined(X_LOCALE))
 
 #ifndef SEEK_SET
 # define SEEK_SET 0
@@ -2477,7 +2497,7 @@ done:
     return buffer;
 }
 
-# if defined(FEAT_EVAL) || defined(PROTO)
+# if defined(FEAT_EVAL)
 
     static void
 get_cmd_output_as_rettv(
@@ -2782,7 +2802,7 @@ path_with_url(char_u *fname)
     return path_is_url(p);
 }
 
-#if defined(FEAT_EVAL) || defined(PROTO)
+#if defined(FEAT_EVAL)
 /*
  * Return the dictionary of v:event.
  * Save and clear the value in case it already has items.
